import os.path
import urllib.request

from anylogiccloudclient.client.cloud_error import CloudError
from anylogiccloudclient.client.http_client import HttpClient
from anylogiccloudclient.client.utils import case_insensitive_equals
from anylogiccloudclient.client.utils import compute_file_hash
from anylogiccloudclient.client.inputs import Inputs

from anylogiccloudclient.data.model import Model
from anylogiccloudclient.client.model_run import ModelRun
from anylogiccloudclient.data.model_version import ModelVersion
from anylogiccloudclient.data.file_resource import FileResource
from io import IOBase
from http.client import HTTPResponse


class CloudClient:
    VERSION = "8.5.0"

    def __init__(self, api_key: str, host="https://cloud.anylogic.com"):
        self.api_key = api_key
        self._set_host(host)
        self._http_client = HttpClient(self.root_url, api_key)

    def _set_host(self, host) -> None:
        self.host_url = host
        self.root_url = self.host_url + "/api/open/" + self.VERSION + "/"

    def get_models(self) -> list:
        response = self._http_client.api_request("models")
        models = list(map(lambda m: Model.from_json(m), response))
        return models

    def get_model_by_id(self, id: str) -> Model:
        response = self._http_client.api_request("models/" + id)
        model = Model.from_json(response)
        return model

    def get_model_by_name(self, name: str) -> Model:
        response = self._http_client.api_request("models/name", "POST", name)
        model = Model.from_json(response)
        return model

    def get_model_version_by_id(self, model: Model, version_id: str) -> ModelVersion:
        response = self._http_client.api_request("models/" + model.id + "/versions/" + version_id)
        version = ModelVersion.from_json(response)
        return version

    def get_model_version_by_number(self, model: Model, version_number: int) -> ModelVersion:
        response = self._http_client.api_request("models/" + model.id + "/versions/number/" + str(version_number))
        version = ModelVersion.from_json(response)
        return version

    def get_latest_model_version(self, model: Model or str) -> ModelVersion:
        if isinstance(model, str):
            model = self.get_model_by_name(model)
        version = self.get_model_version_by_id(model, model.model_versions[len(model.model_versions) - 1])
        return version

    def create_default_inputs(self, version: ModelVersion):
        return Inputs(self._http_client, version)

    def create_inputs_from_experiment(self, version: ModelVersion, experiment_name: str):
        experiments = self._get_model_version_experiments(version)
        try:
            experiment = next(e for e in experiments if case_insensitive_equals(e["name"], experiment_name))
        except StopIteration as e:
            raise CloudError("Experiment with name: '" + experiment_name + "' not found")
        return Inputs(self._http_client, version, experiment)

    def create_simulation(self, inputs) -> ModelRun:
        return ModelRun(self._http_client, inputs, inputs._model_version, "SIMULATION")

    def create_parameter_variation(self, inputs) -> ModelRun:
        return ModelRun(self._http_client, inputs, inputs._model_version, "PARAMETER_VARIATION")

    def create_parameter_variation_with_replications(self, inputs) -> ModelRun:
        return ModelRun(self._http_client, inputs, inputs._model_version, "PARAMETER_VARIATION_WITH_REPLICATIONS")

    def create_monte_carlo(self, inputs):
        return ModelRun(self._http_client, inputs, inputs._model_version, "MONTE_CARLO")

    def _get_model_version_experiments(self, model_version):
        return self._http_client.api_request("versions/" + model_version.id + "/experiments")

    def upload_file(self, file) -> str:
        if isinstance(file, str):
            with open(file, "rb") as reader:
                return self._upload_file(reader)
        elif isinstance(file, IOBase):
            return self._upload_file(file)
        else:
            raise CloudError("Unsupported type of 'file' argument")

    def _upload_file(self, stream) -> str:
        content = stream.read()
        if content == b'':
            # Hack: Some implementations of BufferedIOBase (google drive client) return empty string from read()
            # we don't want the user to bother with that so we try a different method that might work
            # but there are also implementations that do not provide getvalue() (http.client.HTTPResponse)
            # so we check if these is a getvalue() method first.
            getvalue = getattr(stream, "getvalue", None)
            if callable(getvalue):
                content = stream.getvalue()
        hash = self._get_file_hash(content)
        if not self.file_exists_by_hash(hash):
            return self._http_client.request("upload-file", "POST", content)
        return hash

    def download_file(self, file_resource: FileResource, file=None):
        url = self._http_client.request("output-file?hash=" + file_resource.hash)
        response = urllib.request.urlopen(url)
        if file is None:
            file = file_resource.name
        if isinstance(file, str):
            directory = os.path.dirname(file)
            if directory != "":
                os.makedirs(directory, exist_ok=True)
            with open(file, "wb") as writer:
                writer.write(response.read())
        elif isinstance(file, IOBase):
            file.write(response.read())
        else:
            raise CloudError("Unsupported type of 'file' argument")

    def file_exists_by_hash(self, hash_code: str) -> bool:
        return self._http_client.request("check-file?hash=" + hash_code)

    def get_file_hash(self, file) -> str:
        if isinstance(file, str):
            with open(file, "rb") as reader:
                return self._get_file_hash(reader.read())
        else:
            raise CloudError("Unsupported type of 'file' argument")

    def _get_file_hash(self, str) -> str:
        return compute_file_hash(str)
