import json
import time

from anylogiccloudclient.client.cloud_error import CloudError
from anylogiccloudclient.client.multi_run_outputs import MultiRunOutputs
from anylogiccloudclient.client.single_run_outputs import SingleRunOutputs
from anylogiccloudclient.client.utils import case_insensitive_equals
from anylogiccloudclient.data.aggregation import Aggregation
from anylogiccloudclient.data.aggregation_types import AggregationType


class ModelRun:
    def __init__(self, http_client, inputs, model_version, type):
        self.http_client = http_client
        self.inputs = inputs._clone()
        self.model_version = model_version
        self.type = type
        self.versions_url = "/versions/" + self.model_version.id
        self._run_state = None
        if self.type == "PARAMETER_VARIATION_WITH_REPLICATIONS" and \
                not self._is_in_list("{NUMBER_OF_REPLICATIONS}", self.inputs.names()):
            self.inputs.set_number_of_replications(3)

    def run(self) -> 'ModelRun':
        self.http_client.api_request(self.versions_url + "/runs", "POST", self.inputs._get_key_data(self.type))
        return self

    def wait_for_completion(self, polling_period: int = 5000) -> 'ModelRun':
        self._poll_status(polling_period)
        return self

    def get_outputs(self, required_output_names=None, aggregation_type: AggregationType = None):
        return self._get_run_results(required_output_names, aggregation_type)

    def get_outputs_and_run_if_absent(self, required_output_names=None, aggregation_type: AggregationType = None,
                                      polling_period: int = 5000):
        try:
            return self.get_outputs(required_output_names, aggregation_type)
        except CloudError as e:
            if e.status == 404:
                self.run()
                self.wait_for_completion(polling_period)
                return self.get_outputs(required_output_names, aggregation_type)
            else:
                raise e

    def get_status(self):
        if self._run_state is None:
            self._update_run_state()
        return self._run_state["status"]

    def get_progress(self):
        if self._run_state is None:
            self._update_run_state()
        return json.loads(self._run_state["message"]) if self._run_state["message"] != "" else None

    def stop(self):
        return self.http_client.api_request(self.versions_url + "/runs/stop", "POST",
                                            self.inputs._get_key_data(self.type))

    def _poll_status(self, polling_period):
        while True:
            self._update_run_state()
            if self._run_state["status"] in ["FRESH", "RUNNING"]:
                time.sleep(polling_period / 1000.)
            elif self._run_state["status"] == "COMPLETED":
                return self
            else:
                status = self._run_state["status"]
                message = self._run_state["message"]
                message = message if message is None else message.replace("\\n", "\n")
                raise CloudError(f"status: {status}, message: {message}")

    def _update_run_state(self):
        self._run_state = self.http_client.api_request(self.versions_url + "/run", "POST",
                                                       self.inputs._get_key_data(self.type))

    def _get_run_results(self, required_output_names, aggregation_type):
        if self.type == "SIMULATION":
            return self._get_simulation_run_results(required_output_names)
        elif self.type in ["PARAMETER_VARIATION", "PARAMETER_VARIATION_WITH_REPLICATIONS"]:
            return self._get_variation_run_results(required_output_names, AggregationType.ARRAY)
        elif self.type == "MONTE_CARLO":
            return self._get_monte_carlo_run_results(required_output_names, aggregation_type)

    def _get_simulation_run_results(self, required_output_names):
        aggregations = list(
            map(lambda o: self._create_aggregation(AggregationType.IDENTITY, outputs=[o]),
                self._get_required_outputs(required_output_names))
        )
        result = self._make_request_for_results(aggregations)
        return SingleRunOutputs(result)

    def _get_variation_run_results(self, required_output_names, aggregation_type: AggregationType):
        aggregation_type = aggregation_type or AggregationType.ARRAY
        aggregations = self._create_output_aggregations(aggregation_type, required_output_names)
        variable_inputs = list(filter(lambda i: self._is_range_input(i), self.inputs.inputs_array))
        aggregations.append(self._create_aggregation(aggregation_type, inputs=variable_inputs))
        result = self._make_request_for_results(aggregations)
        return MultiRunOutputs(result)

    def _get_monte_carlo_run_results(self, required_output_names, aggregation_type: AggregationType):
        aggregation_type = aggregation_type or AggregationType.HISTOGRAM_DATA
        aggregations = self._create_output_aggregations(aggregation_type, required_output_names)
        result = self._make_request_for_results(aggregations)
        return MultiRunOutputs(result)

    def _create_output_aggregations(self, aggregation_type, output_names):
        aggregations = list(
            map(lambda o: self._create_aggregation(aggregation_type, outputs=[o]),
                self._get_required_outputs(output_names))
        )
        return aggregations

    def _make_request_for_results(self, aggregations):
        if self._run_state is not None:
            results = self.http_client.api_request(self.versions_url + "/results/" + self._run_state["id"],
                                                   "POST", aggregations)
        else:
            results_request = self.inputs._get_key_data(self.type)
            results_request["outputs"] = json.dumps(aggregations)
            results = self.http_client.api_request(self.versions_url + "/results", "POST", results_request)
        return list(map(lambda r: Aggregation.from_json(r), results))

    def _get_required_outputs(self, required_output_names):
        if required_output_names is not None:
            return filter(lambda o: self._is_in_list(o["name"], required_output_names), self.inputs.outputs)
        else:
            if self.type == "SIMULATION":
                return self.inputs.outputs
            elif self.type in ["PARAMETER_VARIATION", "PARAMETER_VARIATION_WITH_REPLICATIONS"]:
                return list(filter(lambda o: self._is_scalar_type(o["type"]), self.inputs.outputs))
            elif self.type == "MONTE_CARLO":
                return list(filter(lambda o: self._is_histogram_type(o["type"]), self.inputs.outputs))

    def _is_in_list(self, name, list):
        return any(filter(lambda n: case_insensitive_equals(n, name), list))

    def _is_scalar_type(self, type):
        return type in ["BOOLEAN", "INTEGER", "LONG", "DOUBLE", "STRING", "DATE_TIME"]

    def _is_histogram_type(self, type):
        return type in ["INTEGER", "DOUBLE", "DATA_SET"]

    def _is_random_input(self, input):
        return input["type"] in ["RANDOM_INTEGER", "RANDOM_DOUBLE"]

    def _is_range_input(self, input):
        return input["type"] in ["FIXED_RANGE_INTEGER", "FIXED_RANGE_DOUBLE"]

    def _create_aggregation(self, type, inputs=[], outputs=[]):
        return {
            "aggregationType": type.name,
            "inputs": inputs,
            "outputs": outputs
        }
